from pathlib import Path

import numpy as np
from SleePyPhases.PSGEventManager import PSGEventManager
from pyPhases import Phase
from pyPhasesRecordloader import RecordLoader, RecordSignal, Signal


from SleePyPhases.SignalPreprocessing import SignalPreprocessing
from SleePyPhases.phases.Extract import RecordProcessor
import pandas as pd


class DataReport(Phase):
    exampleCount = 5
    exampleSeed = 5
    examplePaddingInS = 2
    exampleSecondsPerInch = 2
    highlight = None
    exampleSkipClass = []
    exampleRawMaxSamplingRate = 256

    def createReportFolder(self, path):
        if not Path(path).exists():
            Path(path).mkdir(parents=True, exist_ok=True)

    def plotRecordSignal(self, classExamples, recordSignal: RecordSignal, fileName, sliceValues=None, highlights=None):
        labelNames = self.getConfig("classification.labelNames")
        classificationName = self.getConfig("classification.name")
        labelFrequency = self.getConfig("preprocessing.labelFrequency")
        samplingrate = recordSignal.targetFrequency
        split = self.getConfig("datasetSplit")

        for classIndex, sliceValues in classExamples.items():
            className = labelNames[classIndex]
            if classIndex in self.exampleSkipClass or className in self.exampleSkipClass:
                continue
            paddingInS = self.examplePaddingInS
            padding = paddingInS * samplingrate
            highlights = []

            factor = samplingrate / labelFrequency
            sliceValues = int(sliceValues[0] * factor), int(sliceValues[1] * factor)

            if self.highlight is not None:
                s, e = self.highlight
                s += paddingInS
                e = e + paddingInS if e > 0 else e - paddingInS
                highlights.append((s, e))

            self.plot.plotRecordSignal(
                recordSignal,
                sliceValues,
                resample=False,
                padding=padding,
                secondsPerInch=self.exampleSecondsPerInch,
                highlights=highlights,
                title="Example for %s/%s in record '{record}' from {from}" % (classificationName, className),
            )
            self.plot.save(f"{fileName}-{split}-{className}")

    def plotExamples(self, stats: dict, name=""):
        currentSplit = self.getConfig("datasetSplit")
        channelNames = [c[0] for c in self.getConfig("preprocessing.targetChannels")]
        classificationName = self.getConfig("classification.name")

        recordStats = stats["records"]
        exampleCount = min(self.exampleCount, len(recordStats))
        # pick random records
        randomRecordIndexes = np.random.choice(len(recordStats), exampleCount, replace=False)

        dm = self.getData("dataversionmanager")
        recordIds = dm.getRecordsForSplit(currentSplit)

        memmapOptions = {
            "dtype": self.getConfig("preprocessing.dtype"),
        }
        dataExporterSignals = self.project.getData("data-processed", np.memmap, options=memmapOptions)
        dataExporterFeatures = self.project.getData("data-features", np.memmap, options=memmapOptions)
        # class example are in the preprocessed label channel resolution
        labelFrequency = self.getConfig("preprocessing.labelFrequency")

        for i, recordIndex in enumerate(randomRecordIndexes):
            # create recordsignal for processed data
            samplingrate = self.getConfig("preprocessing.targetFrequency")
            X, Y = dataExporterSignals[recordIndex], dataExporterFeatures[recordIndex]
            classExamples = recordStats[recordIndex]["classExamples"]
            recordSignal = RecordSignal.fromArray(
                X, sourceFrequency=samplingrate, targetFrequency=samplingrate, names=channelNames, transpose=True
            )
            preprocessing = SignalPreprocessing(self.project.config["preprocessing"])
            labelSignal = Signal(classificationName, Y[:, 0], labelFrequency)
            preprocessing.resample(labelSignal, recordSignal, targetFrequency=samplingrate)
            recordSignal.addSignal(labelSignal)
            recordSignal.recordId = recordIds[recordIndex]

            self.plotRecordSignal(classExamples, recordSignal, f"example-{i}-processed")

            # create recordsignal for raw data
            recordId = recordIds[recordIndex]
            recordLoader = RecordLoader.get()
            recordSignal, events = recordLoader.loadRecord(recordId)
            useSPT = self.getConfig("preprocessing.cutFirstAndLastWake")
            events = RecordProcessor.tailorToSleepScoring(recordSignal, events, useSPT=useSPT)

            samplingrate = int(max(s.frequency for s in recordSignal.signals))
            samplingrate = min(samplingrate, self.exampleRawMaxSamplingRate)
            recordSignal.targetFrequency = samplingrate

            for s in recordSignal.signals:
                if s.frequency != samplingrate:
                    self.logWarning("Resampled signal %s from %i to %i" % (s.name, s.frequency, samplingrate))
                    preprocessing.resample(s, recordSignal, targetFrequency=samplingrate)

            # create and and add label signal
            em = PSGEventManager()
            signalLength = recordSignal.getShape()[1]
            eventSignals = em.getEventSignalFromList(
                events,
                signalLength,
                targetFrequency=samplingrate,
                forceGapBetweenEvents=False,
            )
            labelSignal = Signal(classificationName, eventSignals[classificationName], samplingrate)
            recordSignal.addSignal(labelSignal)

            self.plotRecordSignal(classExamples, recordSignal, f"example-{i}-raw")

    def plotClassDistributions(self, df):
        checkMetaColumns = ["psg_type", "rls_plmd", "insomie", "hypersomnie", "rbd", "sbas_nachrdi_lt5_lt15"]
        for col in checkMetaColumns:
            if col in df.columns:
                self.plot.plotClassDistribution(df.groupby(col)["recordId"].count())
                self.plot.save(name=f"distribution_{col}")

    def main(self):
        from SleePyPhases.Plot import Plot
        
        modelConfigString = self.project.getDataFromName("dataset-bySplit").getTagString()
        self.reportFolder = self.getConfig("reportfolder") + "/" + modelConfigString
        self.createReportFolder(self.reportFolder)
        self.setConfig("datasetSplit", "test")

        analyseDatasets = self.getConfig("datasetSplits")
        self.plot = Plot(self.reportFolder)
        np.random.seed(self.exampleSeed)

        dm = self.getData("dataversionmanager")
        df = pd.DataFrame(self.getData("metadata", list))

        df = df.fillna(
            {
                "rls_plmd": "None",
                "hypersomnie": "None",
                "insomie": "None",
                "rbd": "None",
                "sbas_nachrdi_lt5_lt15": "None",
                "diagnose_rdi_lt5_lt15_nurdiag": "None",
            }
        )

        for datasetName in analyseDatasets:
            self.setConfig("datasetSplit", datasetName)
            stats = self.getData("datasetstats", dict)
            labelNames = stats["labels"]
            classCountDict = dict(zip(labelNames, stats["classCounts"]))

            self.plot.plotClassDistribution(classCountDict, title=f"{datasetName} - Claas Distribution")
            self.plot.save(name=f"{datasetName}-classCounts")

            # plot classification examples
            self.plotExamples(stats)

            # plot metadata distribution
            records = dm.getRecordsForSplit(datasetName)
            self.plotClassDistributions(df.query("recordId in %s" % records))

        self.plotClassDistributions(df)

        testIds = dm.getRecordsForSplit("test")
        testRecords = df.query("recordId in %s" % testIds)

        recordIdsAll = dm.getRecordsForSplit("test") + dm.getRecordsForSplit("trainval")
        allRecords = df.query("recordId in %s" % recordIdsAll)

        stats = ["gender"]

        for r in [allRecords, testRecords]:
            if "gender" in r.columns:
                stats = {
                    "femalePercentage": len(r.query("gender == 'female'")) / len(r),
                }
                dataPoints = ["age", "indexArousal", "countArousal", "ahi", "tst"]
                for d in dataPoints:
                    stats[d] = "%f2±%s" % (r[d].mean(), r[d].std())
                    print("%s: %s" % (d, stats[d]))
                print(stats)



import pandas as pd
from pyPhasesRecordloader import RecordLoader

from pyPhases import Phase

class GatherMetadata(Phase):
    """
    gather metadata
    """
    def main(self):
        rl = RecordLoader.get()
        meta = rl.getAllMetaData()
        pd.DataFrame(meta).to_csv("metadata.csv")


import contextlib
from unittest.mock import MagicMock
import numpy as np
import numpy.testing as npt
import pandas as pd
from pyPhases.test.Mocks import OverwriteConfig
from pyPhases.test.TestCase import TestCase
from pyPhasesRecordloader import RecordSignal, Signal, RecordLoader, SignalPreprocessing
from pyPhasesML.exporter.MemmapRecordExporter import MemmapRecordExporter

from SleePyPhases.phases.BuildDataset import BuildDataset


class TestBuildDatasetSegments(TestCase):
    """ Test Scenario:
        - Dataset with 10 Recordings: training [id 0-3], validation [id 4-6], test [id 7-9]
        - Each Recording just consist of a single channels with incread segment count: [[0], [0, 1], [0, 1, 2], ... , [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]]

    """
    phase = BuildDataset()

    C1 = np.array([1, 2, 3, 4, 5])
    C2 = np.array([6, 7, 8, 9, 10])
    C3 = C2

    def config(self):
        self.project.setConfig("dataversion.split", {})
        return {
            "useLoader": "test",
            "dataversion": {
                "split": {
                    "training": ["0:4"],
                    "validation": ["4:7"],
                    "test": ["7:10"],
                },
                "seed": None,
                "recordIds": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
                "groupBy": None,
            },
            "recordWise": False,
            "segmentLength": 2,
            "segmentLengthLabel": 1,
            "segmentManipulation": [{"name": "addBatchDimension"}],
            "segmentManipulationEval": [{"name": "addBatchDimension"}],
            # "segmentManipulation": [{"name": "fixedSize", "position": "center", "size": 10}],
            # "segmentManipulationEval": [{"name": "fixedSize", "position": "center", "size": 10}],
            # "evalOn": None,
        }
    
    def _createMemmapRecordExporterFromArray(self, array, name):
        length = sum([sum(a) for a in array])
        lengths = [int(len(a)) for a in array]
        path = f"data/test-{name}.memmap"
        shape = (1, sum(lengths), 1)

        memmap = np.memmap(path, dtype="float16", mode="w+", shape=shape)
        start = 0
        for a in array:
            start = start
            end = start + len(a)
            memmap[0, start:end, 0:1] = a
            start = end
        
        dataMemmap = MemmapRecordExporter()
        dataMemmap.recordLengths = lengths
        dataMemmap.recordLengths_cumulative = np.cumsum([0] + lengths)
        dataMemmap.fileShape = tuple(shape)
        dataMemmap.type = "float16"
        dataMemmap.CurrentItemIndex = 0
        dataMemmap.filePath = path

        return dataMemmap

    def setUp(self):
        super().setUp()

        class MyPreprocessing(SignalPreprocessing):
            def add1(signal: Signal, recordSignal: RecordSignal, config: dict):
                signal.signal += 1

        self.preprocessing = MyPreprocessing(self.getConfig("preprocessing"))

        dataProcessed = [np.arange(2*i).reshape(-1, 1) for i in range(1, 11)]
        dataFeatures = [np.arange(i).reshape(i, 1) for i in range(1, 11)]

        self.project.registerData("metadata", pd.DataFrame([{"recordId": str(r)} for r in range(10)]))
        RecordLoader.getRecordList = MagicMock(return_value=["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"])


        self.project.registerData("data-processed", self._createMemmapRecordExporterFromArray(dataProcessed, "data-processed"))
        self.project.registerData("data-features", self._createMemmapRecordExporterFromArray(dataFeatures, "data-features"))

        with self.project:
            self.project.config.update({"dataversion": {"split": None}, "validationSplit": 0.4, "testSplit": 0.3})
            self.project.registerData("data-processed", self._createMemmapRecordExporterFromArray(dataProcessed, "data-processed"))
            self.project.registerData("data-features", self._createMemmapRecordExporterFromArray(dataFeatures, "data-features"))

    def _validateRecord(self, index, data):
        for i in range(index + 1):
            x, y = data.__next__()
            npt.assert_equal(x[0, :, 0], [i*2, i*2+1])
            npt.assert_equal(y[0, :, 0], [i])

    
    @OverwriteConfig({"recordWise": True, "dataversion": {"split": None}, "validationSplit": 0.4, "testSplit": 0.3})
    def testRecordOverview(self):
        data = self.getData("dataset-test")
        npt.assert_equal(data[0][0], np.array([0, 1]).reshape(1, -1, 1))
        npt.assert_equal(data[0][1], np.array([0]).reshape(1, -1, 1))
        
        npt.assert_equal(data[1][0], np.array([0, 1, 2, 3]).reshape(1, -1, 1))
        npt.assert_equal(data[1][1], np.array([0, 1]).reshape(1, -1, 1))

        npt.assert_equal(data[2][0], np.array([0, 1, 2, 3, 4, 5]).reshape(1, -1, 1))
        npt.assert_equal(data[2][1], np.array([0, 1, 2]).reshape(1, -1, 1))

        data = self.getData("dataset-validation")

        npt.assert_equal(data[0][0], np.array([0, 1, 2, 3, 4, 5, 6, 7]).reshape(1, -1, 1))
        npt.assert_equal(data[0][1], np.array([0, 1, 2, 3]).reshape(1, -1, 1))
    
        npt.assert_equal(data[1][0], np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).reshape(1, -1, 1))
        npt.assert_equal(data[1][1], np.array([0, 1, 2, 3, 4]).reshape(1, -1, 1))

        npt.assert_equal(data[2][0], np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]).reshape(1, -1, 1))
        npt.assert_equal(data[2][1], np.array([0, 1, 2, 3, 4, 5]).reshape(1, -1, 1))

        data = self.getData("dataset-training")

        npt.assert_equal(data[0][0], np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13]).reshape(1, -1, 1))
        npt.assert_equal(data[0][1], np.array([0, 1, 2, 3, 4, 5, 6]).reshape(1, -1, 1))

        npt.assert_equal(data[1][0], np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]).reshape(1, -1, 1))
        npt.assert_equal(data[1][1], np.array([0, 1, 2, 3, 4, 5, 6, 7]).reshape(1, -1, 1))

        npt.assert_equal(data[2][0], np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17]).reshape(1, -1, 1))
        npt.assert_equal(data[2][1], np.array([0, 1, 2, 3, 4, 5, 6, 7, 8]).reshape(1, -1, 1))

        npt.assert_equal(data[3][0], np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19]).reshape(1, -1, 1))
        npt.assert_equal(data[3][1], np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9]).reshape(1, -1, 1))
    

    def testDataTraining(self):
        data = self.getData("dataset-training")

        # segmentCount: 1 + 2 + 3 + 4 = 10
        self.assertEqual(len(data), 10)
        dataIter = iter(data)

        self._validateRecord(0, dataIter)
        self._validateRecord(1, dataIter)
        self._validateRecord(2, dataIter)
        self._validateRecord(3, dataIter)

    def testDataValidation(self):
        data = self.getData("dataset-validation")

        # segmentCount: 5 + 6 + 7 = 18
        self.assertEqual(len(data), 18)
        dataIter = iter(data)

        self._validateRecord(4, dataIter)
        self._validateRecord(5, dataIter)
        self._validateRecord(6, dataIter)

    def testDataTest(self):
        data = self.getData("dataset-test")

        # segmentCount: 8 + 9 + 10 = 27
        self.assertEqual(len(data), 27)
        dataIter = iter(data)

        self._validateRecord(7, dataIter)
        self._validateRecord(8, dataIter)
        self._validateRecord(9, dataIter)


    @OverwriteConfig({"BuildDataset": {"useMultiThreading": True, "threads": 2}})
    def testValidationMultiThreading(self):
        with contextlib.suppress(KeyError):
            self.project.unregister("dataset-validation")
        self.testDataValidation()

    @OverwriteConfig({"BuildDataset": {"useMultiThreading": True, "threads": 2}})
    def testTrainingMultiThreading(self):
        with contextlib.suppress(KeyError):
            self.project.unregister("dataset-training")
        self.testDataTraining()

    
    @OverwriteConfig({"dataversion": {"split": None}, "validationSplit": 0.4, "testSplit": 0.3})
    def testSingleSegment(self):
        data = self.getData("dataset-test")

        # first segment (record 0 segment 0)
        X, Y = data[0]
        npt.assert_equal(X[0], np.array([0, 1]).reshape(-1, 1))
        npt.assert_equal(Y[0], np.array([0]).reshape(-1, 1))

        # 3nd segment (record 1 segment 1)
        X, Y = data[1]
        npt.assert_equal(X[0], np.array([0, 1]).reshape(-1, 1))
        npt.assert_equal(Y[0], np.array([0]).reshape(-1, 1))

        # 4th segment (record 1 segment 2)
        X, Y = data[2]
        npt.assert_equal(X[0], np.array([2, 3]).reshape(-1, 1))
        npt.assert_equal(Y[0], np.array([1]).reshape(-1, 1))

        # 8th segment (record 3 segment 1)
        X, Y = data[5]
        npt.assert_equal(X[0], np.array([4, 5]).reshape(-1, 1))
        npt.assert_equal(Y[0], np.array([2]).reshape(-1, 1))

    @OverwriteConfig({"dataversion": {"split": None}, "validationSplit": 0.4, "testSplit": 0.3})
    def testAutoSplit(self):
        with contextlib.suppress(KeyError):
            self.project.unregister("dataset-training")
            self.project.unregister("dataset-validation")
            self.project.unregister("dataset-test")

            
        data = self.getData("dataset-test")

        # segmentCount: 1 + 2 + 3 = 6
        self.assertEqual(len(data), 6)
        dataIter = iter(data)

        self._validateRecord(0, dataIter)
        self._validateRecord(1, dataIter)
        self._validateRecord(2, dataIter)

        data = self.getData("dataset-validation")

        # segmentCount: 4 + 5 + 6 = 15
        self.assertEqual(len(data), 15)
        dataIter = iter(data)

        self._validateRecord(3, dataIter)
        self._validateRecord(4, dataIter)
        self._validateRecord(5, dataIter)
        
        data = self.getData("dataset-training")

        # segmentCount: 7 + 8 + 9 + 10 = 34
        self.assertEqual(len(data), 34)
        dataIter = iter(data)

        self._validateRecord(6, dataIter)
        self._validateRecord(7, dataIter)
        self._validateRecord(8, dataIter)
        self._validateRecord(9, dataIter)


import numpy as np
from pyPhases.test.Mocks import OverwriteConfig
from pyPhases.test.TestCase import TestCase
from pyPhasesML import DataversionManager

from SleePyPhases.phases.Setup import Setup


class TestSetup(TestCase):
    phase = Setup()
    
    def config(self):
        self.project.setConfig("dataversion.split", {})
        return {
            "fold": 0,
            "validationSplit": 0.25,
            "dataversion": {
                "seed": None,
                "split": {
                    "trainvaltest": ["0:10"],
                },
                "folds": 5,
                "groupBy": None,
                "recordIds": ["0", "1", "2", "3", "4", "5", "6", "7", "8", "9"],
            },
        }
    
    def testConfigValues(self):
        self.assertEqual(self.getConfig("numLabels"), 1)
        self.assertEqual(self.getConfig("numClasses"), [2])
        # self.assertEqual(self.getConfig("fold"), 0)
        # self.assertEqual(self.getConfig("datasetSplits"), [])
        # self.assertEqual(self.getConfig("hasFolds"), False)

    # def testDataVersionManagerTestFold(self):

    #     # self.project.setConfig("dataversion.split", {})
    #     # self.project.config.update("dataversion", {"folds": 5, "split": {"trainvaltest": ["0:10"]}})
    #     dm = self.getData("dataversionmanager", DataversionManager)
    #     self.assertEqual(dm.getRecordsForSplit("test"), ["0", "1"])
    #     self.assertEqual(dm.getRecordsForSplit("validation"), ["2", "3"])
    #     self.assertEqual(dm.getRecordsForSplit("training"), ["4", "5", "6", "7", "8"])

    # @OverwriteConfig({"fold": 1})
    # def testDataVersionManagerTestFold1(self):
    #     # self.project.setConfig("dataversion.split", {})
    #     # self.project.config.update("dataversion", {"folds": 5, "split": {"trainvaltest": ["0:10"]}})
        

    #     dm = self.getData("dataversionmanager", DataversionManager)
    #     self.assertEqual(dm.getRecordsForSplit("test"), ["0", "1"])
    #     self.assertEqual(dm.getRecordsForSplit("validation"), ["2", "3"])
    #     self.assertEqual(dm.getRecordsForSplit("training"), ["4", "5", "6", "7", "8"])



import unittest
from SleePyPhases.SignalPreprocessing import SignalPreprocessing
from pyPhasesRecordloader import RecordSignal, Signal
import numpy as np

class TestSignalPreprocessing(unittest.TestCase):

    def setUp(self):
        self.preprocessor = SignalPreprocessing({"stepsPerType": {}})

        recordSignal = RecordSignal("test_record", 100)
        recordSignal.addSignal(Signal("ch1", np.random.randn(100), frequency=100))
        recordSignal.addSignal(Signal("ch2", np.random.randn(100), frequency=100))
        self.recordSignal = recordSignal

    def test_rls(self):
        
        signal = Signal("test", np.random.randn(100), frequency=100)

        self.preprocessor.rls(signal, self.recordSignal, ["ch1", "ch2"], [])
        
        self.assertIsInstance(signal.signal, np.ndarray)
        self.assertEqual(signal.signal.shape, (100,))
        self.assertFalse(np.any(np.isnan(signal.signal)))

    
    def test_zerophase(self):
        b = np.array([0.1, 0.2, 0.3])
        a = np.array([1.0, -0.3, 0.2])
        x = np.array([1.0, 2.0, 3.0, 4.0])
        
        result = self.preprocessor._zerophase(b, a, x)
        
        # Add your assertions here
        # For example, to check type:
        self.assertIsInstance(result, np.ndarray)
        
        # To check against known values:
        expected_result = np.array([0.60434229, 0.9425683 , 0.535641  , 0.18467   ])  # Replace with the actual expected result
        np.testing.assert_array_almost_equal(result, expected_result, decimal=5)



# from unittest import TestCase

# import numpy as np
# from pyPhasesRecordloader import Event

# from SleePyPhases.PSGEventManager import PSGEventManager
# from SleePyPhases.RecordFeatureExtraction import RecordFeatureExtraction


# class TestRecordFeatureExtraction(TestCase):
#     def test_sleep_cycle_normal(self):
#         stagesEpochs = np.array([0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4])
#         stagesEpochs = np.repeat(stagesEpochs, 30)
#         stagesSecondSignal = np.repeat(stagesEpochs, 30)
#         events = PSGEventManager().getEventsFromSignal(stagesSecondSignal, ["W", "N1", "N2", "N3", "R"])


#         # Process using both methods
#         new_result = RecordFeatureExtraction(None).SleepCycles(None, events, None)
        
#         expected = {'start': [90, 360, 690], 'stop': [270, 600, 1110]}
#         # expected = calculateCycles(stagesEpochs, 0)
#         assert expected == new_result, "Results do not match"

#     def test_sleep_cycle_shortRem(self):
#         stagesEpochs = np.array([0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4])
#         stagesEpochs = np.repeat(stagesEpochs, 5)

#         stagesSecondSignal = np.repeat(stagesEpochs, 30)
#         events = PSGEventManager().getEventsFromSignal(stagesSecondSignal, ["W", "N1", "N2", "N3", "R"])


#         new_result = RecordFeatureExtraction(None).SleepCycles(None, events, None)

#         expected = {'start': [115], 'stop': [185]}
#         # expected = calculateCycles(stagesEpochs, 0)
#         assert expected == new_result, "Results do not match"

#     def test_sleep_cycle_offset(self):
#         stagesEpochs = np.array([0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 4, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4])
#         stagesEpochs = np.repeat(stagesEpochs, 10)
#         stagesSecondSignal = np.repeat(stagesEpochs, 30)
#         events = PSGEventManager().getEventsFromSignal(stagesSecondSignal, ["W", "N1", "N2", "N3", "R"])
#         events.append(Event("lightOff", 100))


#         new_result = RecordFeatureExtraction(None).SleepCycles(None, events, None)

#         expected = {'start': [220, 330], 'stop': [300, 470]}
#         # expected = calculateCycles(stagesEpochs, 100)
#         assert expected == new_result, "Results do not match"



from pathlib import Path
import numpy as np
from pyPhases import Phase
from pyPhases.util import BatchProgress
from pyPhasesRecordloader import AnnotationNotFound, ChannelsNotPresent, ParseError, RecordLoader

from SleePyPhases.FeatureExtraction import FeatureExtraction
from SleePyPhases.PreManipulation import PreManipulation
from SleePyPhases.PSGEventManager import PSGEventManager
from SleePyPhases.SignalPreprocessing import SignalPreprocessing


class RecordProcessor:
    """this class is needed to tailor the multithreading"""

    def __init__(
        self,
        recordLoader,
        preProcessingConfig,
        signalProcessing,
        eventManager,
        labelChannels,
        project=None,
    ) -> None:
        self.recordLoader = recordLoader
        self.preProcessingConfig = preProcessingConfig
        self.signalProcessing = signalProcessing
        self.eventManager = eventManager
        self.labelChannels = labelChannels
        self.channelNotPresent = []
        self.project = project

    def prepareLabelSignal(self, name, eventSignal, signalLength):

        labelSignalArray = np.zeros(signalLength)
    
        if name == "SleepArousals":
            if "arousal" in eventSignal:
                eventSignal = eventSignal["arousal"]
                labelSignalArray[eventSignal > 1] = 1
        elif name == "SleepStagesAASM":
            if "sleepStage" not in eventSignal:
                raise ChannelsNotPresent("sleepStage")
            eventSignal = eventSignal["sleepStage"]
            labelSignalArray[eventSignal == PSGEventManager.INDEX_WAKE] = 0
            labelSignalArray[eventSignal == PSGEventManager.INDEX_REM] = 1
            labelSignalArray[eventSignal == PSGEventManager.INDEX_NREM1] = 2
            labelSignalArray[eventSignal == PSGEventManager.INDEX_NREM2] = 3
            labelSignalArray[eventSignal == PSGEventManager.INDEX_NREM3] = 4
        elif name == "SleepApnea":
            if "apnea" in eventSignal:
                eventSignal = eventSignal["apnea"]
                labelSignalArray[:] = 0
                labelSignalArray[eventSignal == PSGEventManager.INDEX_APNEA_OBSTRUCTIVE] = 1
                labelSignalArray[eventSignal == PSGEventManager.INDEX_APNEA_MIXED] = 2
                labelSignalArray[eventSignal == PSGEventManager.INDEX_APNEA_CENTRAL] = 3
                labelSignalArray[eventSignal == PSGEventManager.INDEX_APNEA_HYPO] = 4
        elif name == "SleepLegMovements":
            if "limb" in eventSignal:
                eventSignal = eventSignal["limb"]
                labelSignalArray[:] = 0
                labelSignalArray[eventSignal > 1] = 1
        elif name == "SleepCombined":
            if "sleepStage" not in eventSignal:
                raise ChannelsNotPresent("sleepStage")
            labelSignalArray = np.zeros((signalLength, 4))
            labelSignalArray[:, 0] = self.prepareLabelSignal("SleepStagesAASM", eventSignal, signalLength)
            labelSignalArray[:, 1] = self.prepareLabelSignal("SleepArousals", eventSignal, signalLength)
            labelSignalArray[:, 2] = self.prepareLabelSignal("SleepApnea", eventSignal, signalLength)
            labelSignalArray[:, 3] = self.prepareLabelSignal("SleepLegMovements", eventSignal, signalLength)
        else:
            raise Exception("unknown label channel: " + name)
        return labelSignalArray

    def preparelabelChannels(self, eventSignal, signalLength):
        labelChannels = []
        for name in self.labelChannels:
            signal = self.prepareLabelSignal(name, eventSignal, signalLength)
            signal = signal.reshape(-1, 1)
            labelChannels.append(signal)

        return np.concatenate(labelChannels, axis=1)

    @staticmethod
    def tailorToSleepScoring(recordSignal, events, useSPT=False, fillLastSleepstage=False, addOffset=None, cutToAnnotation=False):
        sleepEventNames = ["R", "N1", "N2", "N3"]
        allSleepStages = ["W", "undefined"] + sleepEventNames
        offsetStart = None
        offsetEnd = None
        addOffset = addOffset or [0, 0]


        if not useSPT:
            # add wake to relevant stages to cut only undefined
            sleepEventNames = allSleepStages
            ligthEvents = [e for e in events if e.name in ["lightOn", "lightOff"]]
            if lightOff := [e for e in ligthEvents if e.name == "lightOff"]:
                offsetStart = lightOff[0].start

            if lightOn := [e for e in ligthEvents if e.name == "lightOn"]:
                offsetEnd = lightOn[0].start

        sleepEvents = [e for e in events if e.name in sleepEventNames]
        allSleepEvents = [e for e in events if e.name in allSleepStages]

        if not sleepEvents:
            raise ChannelsNotPresent("SleepStagesAASM")

        startOffset = sleepEvents[0].start if offsetStart is None else offsetStart
        if sleepEvents[-1].duration == 0:
            sleepEvents[-1].duration = len(recordSignal) / recordSignal.targetFrequency - sleepEvents[-1].start if fillLastSleepstage else 30
        endOffset = sleepEvents[-1].end() if offsetEnd is None else offsetEnd

        if startOffset == 0 and endOffset is None:
            return events

        startOffset += addOffset[0]
        endOffset += addOffset[1]

        startOffset = max(allSleepEvents[0].start, startOffset)
        endOffset = min(allSleepEvents[-1].end(), endOffset)

        # if the signal is shorter than sleep annotation, cut the events and smooth to 30 seconds
        if cutToAnnotation:
            _, maxSignalLength = recordSignal.getShape()
            if endOffset * recordSignal.targetFrequency > maxSignalLength:
                endOffset = maxSignalLength / recordSignal.targetFrequency
                endOffset = int(endOffset / 30) * 30
        
        recordSignal.signalOffset(startOffset, endOffset)

        newEvents = []
        for event in events:
            if startOffset > 0:
                if event.end() <= startOffset:
                    continue
                if event.start < startOffset:
                    event.duration -= startOffset - event.start
                    event.start = 0
                else:
                    event.start -= startOffset
            if endOffset is not None:
                if event.start >= endOffset - startOffset:
                    continue
                if event.end() > endOffset - startOffset:
                    event.duration -= event.end() - endOffset + startOffset
            newEvents.append(event)

        if not event:
            raise ChannelsNotPresent("Arousals")

        return newEvents

    def __call__(self, recordId):
        try:
            recordSignal, events = self.recordLoader.loadRecord(recordId)

            targetSignals = self.preProcessingConfig["targetChannels"]
            targetFrequency = self.preProcessingConfig["targetFrequency"]
            labelFrequency = self.preProcessingConfig["labelFrequency"]
            featureChannels = self.preProcessingConfig["featureChannels"]
            manipulationSteps = self.preProcessingConfig["manipulationSteps"]
            recordSignal.targetFrequency = targetFrequency

            # set the target frequency for the signal to get the correct signal length
            self.signalProcessing.preprocessingSignal(recordSignal)

            recordSignal, events = PreManipulation.getInstance(manipulationSteps)(recordSignal, events)
            FeatureExtraction.getInstance(self.project).extractChannelsByConfig(recordSignal, featureChannels)

            signalLength = round(recordSignal.getSignalLength() * labelFrequency / targetFrequency)
            eventSignal = self.eventManager.getEventSignalFromList(
                events,
                signalLength,
                targetFrequency=labelFrequency,
                forceGapBetweenEvents=False,
            )
            # signalLength * 30
            eventSignal = self.preparelabelChannels(eventSignal, signalLength)
            processedSignal = recordSignal.getSignalArray(targetSignals)
        except ChannelsNotPresent as e:
            self.channelNotPresent.append(e.channels)
            print(f"\033[31;1;4m%s\033[0mChannel missing in {recordId}: {e.channels} ... skipping record")
            return None
        except AnnotationNotFound as e:
            self.channelNotPresent.append(e.name)
            print(f"\033[31;1;4m%s\033[0mAnnotation missing in {recordId}: {e.name} ... skipping record")

            return None
        except ParseError as e:
            print(f"\033[31;1;4m%s\033[0mParseError in record {recordId}, record will be skipped: {e}")
            return None
        
        return processedSignal, eventSignal


class Extract(Phase):
    """
    Extract numpy arrays from the raw records in the dataset
    """

    useMultiThreading = True
    threads = None

    def getRecordProcessor(self):
        preprocessingConfig = self.getConfig("preprocessing")
        processRecord = RecordProcessor(
            recordLoader=RecordLoader.get(),
            preProcessingConfig=preprocessingConfig,
            signalProcessing=SignalPreprocessing.getInstance(preprocessingConfig),
            eventManager=PSGEventManager(),
            labelChannels=self.getConfig("labelChannels"),
            project=self.project,
        )
        return processRecord

    def extractRecord(self, recordId):
        
        processRecord = self.getRecordProcessor()
        result = processRecord(recordId)      
        if result is None:
            raise ChannelsNotPresent(processRecord.channelNotPresent)
        return result

    def buildFromRecords(self, records, force=False):
        processRecord = self.getRecordProcessor()
        exporterSignals, dataIdSignals = self.project.getExporterAndId("data-processed", np.memmap)
        exporterFeatures, dataIdFeatures = self.project.getExporterAndId("data-features", np.memmap)
        exporterSignals.options["dtype"] = self.getConfig("preprocessing.dtype", "float32")
        exporterFeatures.options["dtype"] = self.getConfig("preprocessing.dtype", "float32")

        removedRecordIds = []
        removed_records_path = exporterSignals.getPath(f"{dataIdSignals}-removed-records-tmp.npy")
        
        if force is False:
            if exporterSignals.exists(dataIdSignals) and exporterFeatures.exists(dataIdFeatures):
                self.logSuccess("Datafiles already exist")
                return []

            if exporterSignals.existsTmp(dataIdSignals) and exporterFeatures.existsTmp(dataIdFeatures):
                exporterSignals.loadTmp(dataIdSignals)
                exporterFeatures.loadTmp(dataIdFeatures)
                
                if Path(removed_records_path).exists():
                    removedRecordIds = np.load(removed_records_path, allow_pickle=True).tolist()

                recordsLoaded = len(exporterFeatures.currentArrayShapes[dataIdSignals][0])
                total_processed = recordsLoaded + len(removedRecordIds)
                
                # Skip both successfully processed and removed records
                records = records[total_processed:]
                self.logSuccess(f"Resuming extraction: {recordsLoaded} records already loaded, {len(removedRecordIds)} records skipped, {len(records)} to go")

        bp = BatchProgress(records)
        bp.useMultiThreading = self.useMultiThreading
        if self.threads is not None:
            bp.threads = self.threads

        def combine(recordList, dataIndex):
            recordList = recordList.copy()
            # append the the data and the shape to a tmp file

            # check if all records where loaded, else extend the removedRecord list
            if None in recordList:
                rm = [i + dataIndex for i, r in enumerate(recordList) if r is None]
                new_removed = [records[r] for r in rm]
                removedRecordIds.extend(new_removed)
                np.save(removed_records_path, removedRecordIds)
                
                recordList = [r for r in recordList if r is not None]
                self.logWarning("Skipped %i records: channels/annotation not present" % len(rm))

            if len(recordList) > 0:
                signals, features = zip(*recordList)
                exporterSignals.saveAndAppendArray(dataIdSignals, signals)
                exporterFeatures.saveAndAppendArray(dataIdFeatures, features)

        bp.start(processRecord, afterBatch=combine)

        # save the tmp files to the final files
        if not exporterSignals.exists(f"{dataIdSignals}-tmp"):
            raise Exception("No record was stored during the extraction, please check logs for skipped records")
        exporterSignals.finishStream(dataIdSignals)
        exporterFeatures.finishStream(dataIdFeatures)

        return removedRecordIds

    def generateData(self, name):
        
        # unregister modelstate in case gpu memory is needed for extraction
        if self.project.dataExists(self.project.getDataFromName("modelState")):
            self.project.unregister("modelState")

        if name == "record-preprocessed":
            recordId = self.getConfig("recordId")
            return self.extractRecord(recordId)
        elif name[:5] == "data-":
            groupedRecords = self.project.getData("allDBRecordIds", list)
            flatten_ids = [r for ids in groupedRecords.values() for r in ids]

            removedRecordIds = self.buildFromRecords(flatten_ids)

            signals = self.getData("data-processed", np.memmap, generate=False)
            features = self.getData("data-features", np.memmap, generate=False)

            self.registerData("data-processed", signals)
            self.registerData("data-features", features)
            self.registerData("removedRecordIds", removedRecordIds)

    def main(self):
        self.generateData("data-processed")



from unittest.mock import Mock
import numpy as np

from pyPhasesRecordloader import RecordSignal, Event, ChannelsNotPresent, Signal
from SleePyPhases.PreManipulation import PreManipulation

class TestPreManipulation:

    def setUp(self):

        # Setup equivalent to the 'record_signal' pytest fixture
        self.mock_signal = RecordSignal(targetFrequency=10)
        self.mock_signal.signals = [Signal("test_1", np.arange(1000), frequency=10)]

    def test_discard_records_not_all_sleepstages_valid(self):
        pre_manip = PreManipulation({})
        events = [
            Event("W", 0, 30),
            Event("N1", 30, 30),
            Event("N2", 60, 30),
            Event("N3", 90, 30),
            Event("R", 120, 30)
        ]
        record_signal = Mock()
        
        result_signal, result_events = pre_manip.discard_records_not_all_sleepstages(record_signal, events)
        assert result_events == events

    def test_discard_records_missing_stage_raises(self):
        pre_manip = PreManipulation({})
        events = [
            Event("W", 0, 30),
            Event("N1", 30, 30),
            Event("N2", 60, 30),
            # Missing N3
            Event("R", 90, 30)
        ]
        record_signal = Mock()
        
        with pytest.raises(ChannelsNotPresent):
            pre_manip.discard_records_not_all_sleepstages(record_signal, events)

    def test_remove_undefined_events(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation({})
        events = [
            Event("W", 0, 30),
            Event("undefined", 30, 30),
            Event("N1", 60, 30)
        ]
        
        result_signal, result_events = pre_manip.remove_undefined(record_signal, events)
        assert len(result_events) == 2
        assert all(e.name != "undefined" for e in result_events)
        assert result_events[-1].start == 30
        assert (result_signal.signals[0].signal[0:300] == np.arange(1000)[0:300]).all()
        assert (result_signal.signals[0].signal[300:] == np.arange(1000)[600:]).all()

    def test_remove_undefined_events_default(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation([{"name": "test_step"}])
        events = [
            Event("W", 0, 30),
            Event("undefined", 30, 30),
            Event("N1", 60, 30)
        ]
        
        result_signal, result_events = pre_manip.remove_undefined(record_signal, events)
        assert len(result_events) == 2
        assert all(e.name != "undefined" for e in result_events)
        assert result_events[-1].start == 30
        assert (result_signal.signals[0].signal[0:300] == np.arange(1000)[0:300]).all()
        assert (result_signal.signals[0].signal[300:] == np.arange(1000)[600:]).all()

    def test_remove_undefined_events_overlap(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation({})
        events = [
            Event("W", 0, 30),
            Event("arousal_start", 25, 10),
            Event("arousal_big", 25, 5+30+5),
            Event("undefined", 30, 30),
            Event("arousal_end", 55, 15),
            Event("N1", 60, 30)
        ]
        
        result_signal, result_events = pre_manip.remove_undefined(record_signal, events)
        assert all(e.name != "undefined" for e in result_events)
        assert result_events[-1].start == 30
        # start
        assert result_events[1].name == "arousal_start"
        assert result_events[1].start == 25
        assert result_events[1].duration == 5
        # end
        assert result_events[3].name == "arousal_end"
        assert result_events[3].start == 30
        assert result_events[3].duration == 10
        # big
        assert result_events[2].name == "arousal_big"
        assert result_events[2].start == 25
        assert result_events[2].duration == 5+5

        assert (result_signal.signals[0].signal[0:300] == np.arange(1000)[0:300]).all()
        assert (result_signal.signals[0].signal[300:] == np.arange(1000)[600:]).all()

    def test_balance_wake_excessive_wake_evening(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation({})
        events = [
            Event("W", 0, 30),  # Excessive wake
            Event("N1", 30, 3),
            Event("N2", 33, 10)
        ]
        
        result_signal, result_events = pre_manip.balance_wake(record_signal, events)
        # start
        assert result_events[0].name == "W"
        assert result_events[0].start == 0
        assert result_events[0].duration == 10
        
        assert (result_signal.signals[0].signal[:] == np.arange(1000)[200:]).all()

    def test_balance_wake_excessive_wake_morning(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation({})
        events = [
            Event("N1", 0, 10),  # Excessive wake
            Event("N2", 10, 10),
            Event("W", 20, 70)
        ]
        
        result_signal, result_events = pre_manip.balance_wake(record_signal, events)
        # start
        assert result_events[-1].name == "W"
        assert result_events[-1].start == 20
        assert result_events[-1].duration == 10
        
        assert (result_signal.signals[0].signal[:] == np.arange(1000)[0:300]).all()

    def test_balance_wake_excessive_wake_both(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation({})
        events = [
            Event("W", 0, 10),
            Event("N2", 10, 10),
            Event("W", 20, 70)
        ]
        
        result_signal, result_events = pre_manip.balance_wake(record_signal, events)
        # start
        assert result_events[0].name == "N2"

        assert result_events[-1].name == "W"
        assert result_events[-1].start == 10
        assert result_events[-1].duration == 10
        
        assert (result_signal.signals[0].signal[:] == np.arange(1000)[100:300]).all()

    def test_balance_wake_excessive_wake_keepmorning(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation({})
        events = [
            Event("W", 0, 10),
            Event("N2", 10, 10),
            Event("W", 20, 20),
            Event("N3", 40, 10),
            Event("W", 50, 30)
        ]
        
        result_signal, result_events = pre_manip.balance_wake(record_signal, events)
        # start
        assert result_events[0].name == "N2"
        assert result_events[0].start == 0
        assert result_events[0].duration == 10

        assert result_events[-1].name == "N3"
        assert result_events[-1].start == 30
        assert result_events[-1].duration == 10
        
        assert (result_signal.signals[0].signal[:] == np.arange(1000)[100:500]).all()

    
    def test_trim_spt_default(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation({})
        events = [
            Event("W", 0, 40),
            Event("N1", 40, 20),
            Event("W", 60, 40)
        ]
        
        result_signal, result_events = pre_manip.trimSPT(record_signal, events, 30, 30)
        assert len(result_events) > 0
        assert result_events[0].name == "W"
        assert result_events[0].start == 0
        assert result_events[0].duration == 30

        assert result_events[1].name == "N1"
        assert result_events[1].start == 30
        assert result_events[1].duration == 20

        assert result_events[2].name == "W"
        assert result_events[2].start == 50
        assert result_events[2].duration == 30

        assert (result_signal.signals[0].signal[:] == np.arange(1000)[100:900]).all()

    def test_trim_spt_evening_only(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation({})
        events = [
            Event("W", 0, 60),
            Event("N1", 60, 30),
            Event("W", 90, 10)
        ]
        
        result_signal, result_events = pre_manip.trimSPT(record_signal, events, 30, 30)
        assert len(result_events) > 0
        assert result_events[0].name == "W"
        assert result_events[0].start == 0
        assert result_events[0].duration == 30
        assert result_events[1].name == "N1"
        assert result_events[1].start == 30
        assert result_events[1].duration == 30
        assert result_events[2].name == "W"
        assert result_events[2].start == 60
        assert result_events[2].duration == 10

        assert (result_signal.signals[0].signal[:] == np.arange(1000)[300:1000]).all()
    
    def test_trim_spt_morning_only(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation({})
        events = [
            Event("W", 0, 30),
            Event("N1", 30, 30),
            Event("W", 60, 40)
        ]
        
        result_signal, result_events = pre_manip.trimSPT(record_signal, events, 30, 30)
        assert len(result_events) > 0
        assert result_events[0].name == "W"
        assert result_events[0].start == 0
        assert result_events[0].duration == 30
        assert result_events[1].name == "N1"
        assert result_events[1].start == 30
        assert result_events[1].duration == 30
        assert result_events[2].name == "W"
        assert result_events[2].start == 60
        assert result_events[2].duration == 30

        assert (result_signal.signals[0].signal[:] == np.arange(1000)[0:900]).all()
    
    def test_trim_spt_no_morning(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation({})
        events = [
            Event("W", 0, 60),
            Event("N1", 60, 30)
        ]
        
        result_signal, result_events = pre_manip.trimSPT(record_signal, events, 30, 30)
        assert len(result_events) > 0
        assert result_events[0].name == "W"
        assert result_events[0].start == 0
        assert result_events[0].duration == 30
        assert result_events[1].name == "N1"
        assert result_events[1].start == 30
        assert result_events[1].duration == 30

        assert (result_signal.signals[0].signal[:] == np.arange(1000)[300:900]).all()
        
    def test_step_invalid_name(self, config):
        pre_manip = PreManipulation({})
        with pytest.raises(Exception) as exc:
            pre_manip.step("invalid_step", Mock(), [])
        assert "PreManipulation 'invalid_step' not found" in str(exc.value)


    def test_trim_to_light_and_scoring(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation({})
        events = [
            Event("W", 0, 20),
            Event("lightOff", 10, 0),
            Event("N1", 20, 30),
            Event("N2", 50, 20),
            Event("lightOn", 80, 0),
            Event("W", 70, 20)
        ]

        result_signal, result_events = pre_manip.trimToLightAndScoring(record_signal, events)

        # Check if trimming is done correctly based on light events
        assert result_events[0].start == 0  # Events should start at 0 after trimming
        assert len(result_events) == 4  # light events removed

        # Verify signal trimming
        assert (result_signal.signals[0].signal[:] == np.arange(1000)[100:800]).all()


    def test_trim_to_light_and_scoring_no_light_events(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation({})
        events = [
            Event("W", 10, 20),
            Event("N1", 30, 30),
            Event("N2", 60, 20),
            Event("W", 80, 20)
        ]
        
        result_signal, result_events = pre_manip.trimToLightAndScoring(record_signal, events)
        
        assert result_events[0].start == 0  # First event should start at 0
        assert result_events[1].start == 20  # First event should start at 0
        assert result_events[2].start == 50  # First event should start at 0
        assert result_events[3].start == 70  # First event should start at 0
        assert len(result_events) == 4  # Should keep all sleep stages
        assert (result_signal.signals[0].signal[:] == np.arange(1000)[100:1000]).all()

    def test_trim_to_light_and_scoring_only_light_off(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation({})
        events = [
            Event("lightOff", 10, 0),
            Event("W", 20, 20),
            Event("N1", 40, 30),
            Event("N2", 70, 20),
            Event("W", 90, 20)
        ]
        
        result_signal, result_events = pre_manip.trimToLightAndScoring(record_signal, events)
        
        assert result_events[0].name == "W"
        assert result_events[0].start == 0
        assert result_events[0].duration == 20
        assert result_events[1].start == 20
        assert result_events[2].start == 50
        assert result_events[3].start == 70
        assert len(result_events) == 4
        assert (result_signal.signals[0].signal[:] == np.arange(1000)[200:]).all()

    def test_trim_to_scoring_with_arousals(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation({})
        events = [
            Event("arousal", 0, 5),
            Event("arousal", 7, 3),
            Event("W", 10, 20),
            Event("N1", 30, 30),
            Event("N2", 60, 20),
            Event("arousal", 70, 10),
            Event("W", 80, 20)
        ]
        
        result_signal, result_events = pre_manip.trimToLightAndScoring(record_signal, events)
        
        assert result_events[0].start == 0
        assert result_events[0].name == "W"
        assert len(result_events) == 5  # Should keep all events including arousal
        assert next(e for e in result_events if e.name == "arousal").start == 60  # Arousal timing adjusted
        assert (result_signal.signals[0].signal[:] == np.arange(1000)[100:1000]).all()
    
    def test_trim_to_light_and_scoring_with_arousals(self):
        record_signal = self.mock_signal
        pre_manip = PreManipulation({})
        events = [
            Event("W", 10, 20),
            Event("N1", 30, 30),
            Event("N2", 60, 20),
            Event("arousal", 70, 10),
            Event("W", 80, 20)
        ]
        
        result_signal, result_events = pre_manip.trimToLightAndScoring(record_signal, events)
        
        assert result_events[0].start == 0
        assert len(result_events) == 5  # Should keep all events including arousal
        assert next(e for e in result_events if e.name == "arousal").start == 60  # Arousal timing adjusted
        assert (result_signal.signals[0].signal[:] == np.arange(1000)[100:1000]).all()